////////////////////////////////////////////////////////////////////////////////
//
// (C) 2010 Sarah Eris Horsley Caffee
//
// mandelbrot - A simple Mandelbrot Set viewer.
//
// App.cpp
// Application class implementation
//
////////////////////////////////////////////////////////////////////////////////

#include <iomanip>
#include <iostream>
#include <sstream>
#include <typeinfo>
#include <vector>

#if defined(WIN32)
#include <cstdlib>
#include <windows.h>
#include <Shlwapi.h>

#else
#include <cctype>
#include <cstring>
#include <cstdio>

#include <dirent.h>
#endif


#include "SDL.h"

#include "App.h"

#include "Options.h"
#include "Mandelbrot.h"
#include "Julia.h"



std::vector<App::view *> App::old_views;

////////////////////////////////////////////////////////////////////////////////

App::App() :
   running (true),
   redraw (false),
   surf_display (NULL),
   surf_selection (NULL),
   surf_fractal (NULL),
   selection_x (-1),
   selection_y (-1),
   selection_width (-1),
   setting_zoom (false),
   can_set_julia_k (false),
   setting_julia_k (false)
   {
   }


////////////////////////////////////////////////////////////////////////////////
App::~App()
   {
   old_views.clear();
   }


////////////////////////////////////////////////////////////////////////////////

int App::run()
   {
   if (init() == false)
      return 1;

   SDL_Event event;

   while (running)
      {
      SDL_WaitEvent(&event);
      do
	 {
	 on_event(&event);
	 }
      while (SDL_PollEvent(&event));

      update();

      if (redraw)
	 render();
      }

   cleanup();

   return 0;
   }


////////////////////////////////////////////////////////////////////////////////
bool App::init()
   {
   if ((SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER)) == -1)
      {
      std::cerr << "SDL_Init() failed: " << SDL_GetError() << std::endl;
      return false;
      }

   if ((surf_display = SDL_SetVideoMode(Options::width, 
					Options::height, 
					Options::bpp, 
					SDL_HWSURFACE | SDL_DOUBLEBUF)) == NULL)
      {
      std::cerr << "SDL_SetVideoMode() failed: " 
		<< SDL_GetError() << std::endl;
      return false;
      }

   SDL_EnableKeyRepeat(500, 100);


   if (!create_new_surface(surf_selection)) { return false; }
   if (!create_new_surface(surf_fractal))   { return false; }


   std::cerr << "Available fractal types are" << std::endl;
   std::map <std::string, FractalCreator *> mymap = get_Fractal_map();
   std::map <std::string, FractalCreator *>::iterator it;
   for ( it=mymap.begin() ; it != mymap.end(); it++ )
      std::cerr << "  " << (*it).first << std::endl;

   FractalCreator * fractal_creator = 
	 get_Fractal_map()[typeid(Mandelbrot).name()];
   if (!fractal_creator)
      {
      std::cerr << "Mandelbrot set not available!" << std::endl;
      return false;
      }
   fractal = fractal_creator->create();
   fractal->init(surf_fractal);
   redraw = true;

   can_set_julia_k = true;


   set_caption();

   SDL_TimerID timer_id = SDL_AddTimer(10, App::timer_callback, NULL);

   return true;
   }


////////////////////////////////////////////////////////////////////////////////

void App::update()
   {
   }


////////////////////////////////////////////////////////////////////////////////

void App::render()
   {
   if (redraw)
      {
      // Start with a clean slate.
      SDL_FillRect(surf_display, &surf_display->clip_rect, 
		   SDL_MapRGB(surf_display->format, 0, 0, 255));

      fractal->calc_set();
      SDL_BlitSurface(surf_fractal, NULL, surf_display, NULL);

      // draw a selection box
      if (setting_zoom)
	 {
	 SDL_BlitSurface(surf_selection, NULL, surf_display, NULL);
	 }

      // Show it to the user.
      SDL_Flip(surf_display);

      redraw = false;
      }
   }


////////////////////////////////////////////////////////////////////////////////

void App::cleanup()
   {

   if (surf_selection)
      {
      SDL_FreeSurface(surf_selection);
      surf_selection = NULL;
      }
   if (surf_display)
      {
      SDL_FreeSurface(surf_display);
      surf_display = NULL;
      }

   SDL_Quit();
   }


////////////////////////////////////////////////////////////////////////////////

void App::on_event(SDL_Event * event)
   {
   EventHandler::on_event(event);
   }


////////////////////////////////////////////////////////////////////////////////

void App::on_KeyDown(SDLKey sym, SDLMod mod, Uint16 unicode)
   {
   switch (sym)
      {
      case SDLK_ESCAPE:
      case SDLK_q:
	 SDL_Event event;
	 event.type = SDL_QUIT;
	 SDL_PushEvent(&event);
	 break;

      case SDLK_b:
      	 restore_view();
	 redraw = true;
	 set_caption();
      	 break;

      case SDLK_s:
      	 save_image();
      	 break;

      case SDLK_UP:
	 fractal->inc_max_iter();
	 set_caption();
	 break;
      case SDLK_DOWN:
	 fractal->dec_max_iter();
	 set_caption();
	 break;


      case SDLK_m:
	 if (fractal)
	    {
	    save_view();
	    delete fractal;
	    FractalCreator * fractal_creator = 
		  get_Fractal_map()[typeid(Mandelbrot).name()];
	    if (!fractal_creator)
	       {
	       std::cerr << "Mandelbrot set not available!" << std::endl;
	       SDL_Event event;
	       event.type = SDL_QUIT;
	       SDL_PushEvent(&event);
	       break;
	       }
	    fractal = fractal_creator->create();
	    fractal->init(surf_fractal);
	    fractal->set_calc_needed();
	    redraw = true;
	    set_caption();
	    can_set_julia_k = true;
	    }
	 break;
      case SDLK_j:
	 if (fractal)
	    {
	    FractalCreator * fractal_creator = 
		  get_Fractal_map()[typeid(Julia).name()];
	    if (!fractal_creator)
	       {
	       std::cerr << "Julia set not available!" << std::endl;
	       }
	    else
	       {
	       save_view();
	       delete fractal;
	       fractal = fractal_creator->create();
	       fractal->init(surf_fractal);
	       fractal->set_calc_needed();
	       redraw = true;
	       set_caption();
	       can_set_julia_k = false;
	       }
	    }
	 break;

      case SDLK_k:
	 if (can_set_julia_k)
	    {
	    if (setting_julia_k) setting_julia_k = false;
	    else setting_julia_k = true;
	    }
	 break;

      default:
	 break;
      }
   }


////////////////////////////////////////////////////////////////////////////////
void App::on_KeyUp(SDLKey sym, SDLMod mod, Uint16 unicode)
   {
   switch (sym)
      {
      case SDLK_UP:
	 fractal->set_calc_needed();
	 redraw = true;
	 break;
      case SDLK_DOWN:
	 fractal->set_calc_needed();
	 redraw = true;
	 break;
      }   
   }


////////////////////////////////////////////////////////////////////////////////

void App::on_Exit()
   {
   running = false;
   }


////////////////////////////////////////////////////////////////////////////////

void App::on_LButtonDown(int mx, int my)
   {
   if (setting_julia_k)
      {
      set_julia_k(mx, my);
      }
   else
      {
      selection_x = mx;
      selection_y = my;
      setting_zoom = true;
      }
   }


////////////////////////////////////////////////////////////////////////////////

void App::on_LButtonUp(int mx, int my)
   {
   if (setting_zoom)
      {
      if ( (mx <= 0) || 
	   (mx >= surf_display->w - 1) || 
	   (my <= 0) || 
	   (my >= surf_display->h - 1))
	 {
	 // Selection cancelled
	 setting_zoom = false;
	 return;
	 }

      selection_width = mouse_move_width(mx, my);

      // calculate new min/max re/im
      double Re_scale = (fractal->get_size())/(surf_fractal->w - 1);
      double Im_scale = (fractal->get_size())/(surf_fractal->h - 1);

      double new_re_min = fractal->get_re_min() + (selection_x) * Re_scale;
      double new_im_max = fractal->get_im_max() - (selection_y) * Im_scale;
      double new_size = selection_width * Re_scale;


      save_view();

      fractal->set_re_min(new_re_min);
      fractal->set_im_max(new_im_max);
      fractal->set_size(new_size);
      fractal->set_calc_needed();
      redraw = true;

      set_caption();

      setting_zoom = false;
      }
   }


////////////////////////////////////////////////////////////////////////////////

void App::on_RButtonDown(int mx, int my)
   {
   if (can_set_julia_k)
      {
      set_julia_k(mx, my);
      }
   }


////////////////////////////////////////////////////////////////////////////////
void App::on_MouseMove(int mx, int my, 
		       int relx, int rely, 
		       bool left, bool right, bool middle)
   {
   if (setting_zoom)
      {
      Uint32 clear = SDL_MapRGBA(surf_selection->format, 
				0, 0, 0, SDL_ALPHA_TRANSPARENT);
      Uint32 white = SDL_MapRGBA(surf_selection->format, 
				 255, 255, 255, SDL_ALPHA_OPAQUE);

      SDL_FillRect(surf_selection, &surf_selection->clip_rect, clear);

      SDL_Rect rect;
      rect.x = selection_x;
      rect.y = selection_y;
      int w = mouse_move_width(mx, my);
      rect.w = rect.h = w;

      SDL_FillRect(surf_selection, &rect, white);

      rect.x = selection_x + 1;
      rect.y = selection_y + 1;
      rect.w = rect.h = w - 2;

      SDL_FillRect(surf_selection, &rect, clear);

      redraw = true;
      }
   }


////////////////////////////////////////////////////////////////////////////////
void App::on_Resize(int w, int h)
   {
   redraw = true;
   }


////////////////////////////////////////////////////////////////////////////////
void App::on_Expose()
   {
   redraw = true;
   }


////////////////////////////////////////////////////////////////////////////////
int App::mouse_move_width(int mx, int my)
   {
   int move_width = (mx - selection_x);
   if (mx > selection_x)
      {
      move_width *= -1;
      }
   
   int tmp_i = (my - selection_y);
   if (my < selection_y)
      {
      tmp_i *= -1;
      }
   
   if (tmp_i > move_width)
      move_width = tmp_i;

   return move_width;
   }


////////////////////////////////////////////////////////////////////////////////
bool App::create_new_surface(SDL_Surface * &surface)
   {
   if (surface)
      {
      SDL_FreeSurface(surface);
      surface = NULL;
      }

   SDL_Surface * tmp;
   if ((tmp = SDL_CreateRGBSurface(SDL_HWSURFACE|SDL_SRCALPHA, 
				   Options::width, Options::height, 
				   Options::bpp, 0, 0, 0, 0))
       == NULL)
      {
      std::cerr << "SDL_CreateRGBSurface() failed in "
		<< "App::create_new_surface(): " 
		<< SDL_GetError() << std::endl;
      return false;
      }

   surface = SDL_DisplayFormatAlpha(tmp);
   SDL_FreeSurface(tmp);

   return true;
   }


////////////////////////////////////////////////////////////////////////////////
void App::set_julia_k(int x, int y)
   {
   save_view();
   
   double k[2];
   k[0]= fractal->get_re_min() + x * 
	 (fractal->get_size())/(surf_fractal->w - 1);
   k[1] = fractal->get_im_max() - y * 
	 (fractal->get_size())/(surf_fractal->h - 1);
      
   FractalCreator * fractal_creator = get_Fractal_map()[typeid(Julia).name()];
   if (!fractal_creator)
      {
      std::cerr << "Julia set not available!" << std::endl;
      return;
      }
   else
      {
      delete fractal;
      fractal = fractal_creator->create();
      fractal->init(surf_fractal);
      fractal->set_option("k", k);
      fractal->set_calc_needed();
      redraw = true;
      
      set_caption();
      setting_julia_k = false;
      }
   }


////////////////////////////////////////////////////////////////////////////////

void App::restore_view()
   {
   if (old_views.size() > 0)
      {
      view * v = old_views.back();
      old_views.pop_back();

      if (v->type.compare(fractal->get_fractal_name()) != 0)
	 {
	 delete fractal;
	 FractalCreator * fractal_creator = get_Fractal_map()[v->type];
	 if (!fractal_creator)
	    {
	    std::cerr << "Cannot restore old view: unknown Fractal type " 
		      << v->type << std::endl;
	    SDL_Event event;
	    event.type = SDL_QUIT;
	    SDL_PushEvent(&event);
	    return;
	    }
	 fractal = fractal_creator->create();
	 fractal->init(surf_fractal);
	 }

      fractal->set_re_min(v->re_min);
      fractal->set_im_max(v->im_max);
      fractal->set_size(v->size);
      fractal->set_max_iter(v->max_iter);
      fractal->set_calc_needed();

      std::cerr << "Restored view:" << std::endl
		<< "  type     " << v->type << std::endl
		<< "  re_min   " << v->re_min << std::endl
		<< "  im_max   " << v->im_max << std::endl
		<< "  size     " << v->size << std::endl
		<< "  max_iter " << v->max_iter << std::endl
		<< std::endl;
      delete v;
      }
   }


////////////////////////////////////////////////////////////////////////////////
void App::save_view()
   {
   view * v = new view;
   v->type = typeid(*fractal).name();
   v->re_min = fractal->get_re_min();
   v->im_max = fractal->get_im_max();
   v->size = fractal->get_size();
   v->max_iter = fractal->get_max_iter();
   old_views.push_back(v);

   std::cerr << "Saved view:" << std::endl
	     << "  type     " << v->type << std::endl
	     << "  re_min   " << v->re_min << std::endl
	     << "  im_max   " << v->im_max << std::endl
	     << "  size     " << v->size << std::endl
	     << "  max_iter " << v->max_iter << std::endl
	     << std::endl;
   }


////////////////////////////////////////////////////////////////////////////////
void App::set_caption()
   {
   std::stringstream ss;
   ss << fractal->get_display_name() << " / "
      << "Real: " << fractal->get_re_min()
       << " to "<< fractal->get_re_min() + fractal->get_size()
      << " / Imaginary: " << fractal->get_im_max() - fractal->get_size()
      << " to " << fractal->get_im_max()
      << " / Max iterations: " << fractal->get_max_iter();
   SDL_WM_SetCaption(ss.str().c_str(), NULL);
   }


////////////////////////////////////////////////////////////////////////////////
bool App::save_image()
   {
   int i = App::get_next_file_num();
   std::stringstream name;
   name << "mandelbrot-" << std::setw(3) << std::setfill('0') << i << ".bmp";

   if (SDL_SaveBMP(surf_display, name.str().c_str()) != 0)
      {
      std::cerr << "Unable to save image file " << name.str() << "" << std::endl;
      return false;
      }

   std::cerr << "Saved image as " << name.str() << std::endl;

   return true;
   }


////////////////////////////////////////////////////////////////////////////////
Uint32 App::timer_callback(Uint32 interval, void *param)
   {
   SDL_Event e;
 
   e.type = SDL_USEREVENT;
   e.user.code = 0;
   e.user.data1 = NULL;
   e.user.data2 = NULL;
   SDL_PushEvent(& e);
 
   return interval;
   }


#if defined(WIN32)
////////////////////////////////////////////////////////////////////////////////
//Code for Windows
int App::get_next_file_num()
   {
   static int i = 0;

   if (i == 0)
      {
      WIN32_FIND_DATA ffd;
      HANDLE hFind = INVALID_HANDLE_VALUE;

      char * file_spec = "mandelbrot-???.bmp";
      hFind = FindFirstFile(file_spec, &ffd);
      if (hFind != INVALID_HANDLE_VALUE)
	 {
	 std::string s;
	 std::stringstream ss;
	 do
	    {
	    std::cerr << ffd.cFileName << std::endl;
	    ss << ffd.cFileName;
	    if (ss.str().size() == 18)
	       {
	       s = ss.str().substr(0, 11);
	       if ( StrCmpI(s.c_str(), "mandelbrot-") == 0 )
		  {
		  s = ss.str().substr(14, 4);
		  if (StrCmpI(s.c_str(), ".bmp") == 0)
		     {
		     s = ss.str().substr(11, 3);
		     int j = atoi(s.c_str());
		     if (j > i) i = j;
		     }
		  }
	       }
	    }
	 while (FindNextFile(hFind, &ffd) != 0);
	 }
      }

   i++;
   return i;
   }



#else
////////////////////////////////////////////////////////////////////////////////
// Code for Linux

int App::scandir_filter(const struct dirent * d)
   {
   if (memcmp(d->d_name, "mandelbrot-", 11) != 0) return 0;
   if (memcmp(d->d_name+14, ".bmp", 4) != 0) return 0;
   if (isdigit(d->d_name[11]) 
       && isdigit(d->d_name[12]) 
       && isdigit(d->d_name[13])) 
      return 1;
   return 0;
   }


int App::get_next_file_num()
   {
   static int i = 0;

   if (i == 0)
      {
      struct dirent ** file_list;
      int num_files = scandir(".", &file_list, scandir_filter, alphasort);
      if (num_files != 0)
	 sscanf(file_list[num_files-1]->d_name, "mandelbrot-%03d.bmp", &i);
      }
   i++;

   return i;
   }

#endif

